---
title: 2-3 tree invariants
nav_order: 3
---

Let's consider a different hypothesis: that unbalanced growth leads to worst-case height trees. Elements are always inserted as new leaf nodes. Sometimes, these new leaves result in unbalanced growth.

If new leaf nodes are responsible for height increase, then suppose we never create new leaves. Instead, when inserting an item, **overstuff** an existing leaf to avoid creating new leaves. If we never create new leaves, the tree can never get unbalanced!

![Overstuffing Leaves](https://docs.google.com/drawings/d/e/2PACX-1vSjmsvafs-zOv3rCyuKifqFfL8C1yPLAaVxFB4_8GOkryAk5wHPCB2bQqnbzspxB49KLBSp4-v33hwP/pub?w=960&h=540)

<details markdown="1">
<summary>Give a worst-case insertion order for overstuffed leaves.</summary>

If we insert items in ascending order, then the leaf will get more and more full. We trade searching *N* children down the side of a spindly tree with searching through *N* items in an overstuffed leaf.
</details>

We can avoid overfull leaves by setting a limit on the number of items in each node. If a node has more than the specified number of items, **promote** an item by pushing it up to its parent node. This needs to be done in a careful manner to maintain the search tree invariant by promoting a middle item and splitting the node into two parts: left and right.

<div style="--aspect-ratio: 720 / 405; --add-height: 29px"><iframe src="https://docs.google.com/presentation/d/e/2PACX-1vSIOhgcqdfEyFYiP-OCC_ZIflWIdfs-6yhth_qte4JC5Cnuczd9m3U3teynutcih85iEPCnHQfJANb1/embed?start=false&loop=false&delayms=3000" frameborder="0" width="100%" height="100%" allowfullscreen="true" mozallowfullscreen="true" webkitallowfullscreen="true"></iframe></div>

A **2-3 search tree** is a balanced search tree data structure where all leaves are the same depth from the root and whose nodes are either:

- **Empty** (null).
- **2-nodes** that contain one key and two children: a left child containing smaller keys and a right child containing larger keys.
- **3-nodes** that contain two keys and three children: a left child containing smaller keys, a middle child containing keys between the node's two keys, and a right child containing larger keys.
