---
title: Proposing invariants
nav_order: 1
---

Goal
: Design a binary tree structure that ensures log height for any insertion order.

Data structure implementers maintain invariants to narrow down exactly where an element can be found in the data structure. Binary search trees maintain the binary search invariant to determine the relative location of elements in the search tree. If we can design an invariant that will determine the shape of the overall tree, then we might be able to guarantee log height for any insertion order.

Let's propose a new invariant by evaluating how we end up with worst-case height trees.

What is the cause of worst-case height trees?
: Worst-case height trees are spindly trees where each node has only one child. In contrast, bushy trees have two children per node.

Design a new invariant that addresses the problem.
: Fullness invariant: each node must have either 0 or 2 children.

<details markdown="1">
<summary>Draw a worst-case binary tree with this invariant.</summary>

![Worst-Case Full Tree](https://docs.google.com/drawings/d/e/2PACX-1vQg6teoRmSeiw10PybR0wxaTVUVr9hW5Lb56BJO48ly2u0KSilPH_BZbH32sXV0DHnchXTCQYD925ZS/pub?w=960&h=540)

The order of growth of the height of this tree is in *Θ*(*N*). We still have a "spindly" structure not because there are no nodes with only 1 child, but because the tree is unbalanced. In practice, we can't actually come up with an insertion order that results in this tree without also inserting nodes with one child. But the bigger problem is that this invariant can still result in linear tree height.
</details>
