---
title: Priority queues
nav_order: 1
---

The **priority queue** is an abstract data type where elements are organized according to their associated priority values. A **max-oriented priority queue** serves highest-priority elements first while a **min-oriented priority queue** serves lowest-priority elements first. A priority queue might be used to help triage emergency room requests. Rather than helping patients first-come, first-served like in a regular queue, we might instead use a priority queue to ensure patients with the most severe ailments (or most time-sensitive needs) are treated first even if someone else arrived before them.

The min-oriented priority queue `MinPQ` interface supports 3 important operations: adding new elements to the priority queue, returning the minimum priority element, and removing the minimum priority element. Unlike sets and maps, **duplicate elements are allowed**, and multiple elements may be added to the priority queue with the same priority value. In case there are two elements with the same priority values, ties can be broken arbitrarily so any element with the minimum priority value could be returned or removed on the next call to those methods. For simplicity, our demos, examples, visualizations in lessons and sections will treat the value of the number or character as its priority.

`add(int element)`
: Add the given `element` to the priority queue.

`min()`
: Returns (without removing) the element with the highest priority.

`removeMin()`
: Removes and returns the element with the highest priority.

We can implement the `MinPQ` interface using data structures that we've already introduced earlier.

`UnorderedLinkedListMinPQ`
: In an unordered linked list, insertion is fast since we can add the element to the front of the list in constant time. Finding the minimum is slow since we need to check the entire linked list.

`OrderedLinkedListMinPQ`
: In an ordered linked list where elements are stored by descending priority value, the opposite is true. Finding the minimum is fast since it's just the first element in the list. Insertion is slow since we need to find the exact position where the element belongs in the ordered list.

`AVLTreeMinPQ`
: A binary search tree organizes elements in sorted order by their priority values. The lowest-priority element is stored in the leftmost node in the tree while the highest-priority element is stored in the rightmost node in the tree.

{% capture question %}
Using intuition, what is the order of growth of the runtime for `AVLTreeMinPQ` operations?

- [ ] *Θ*(1)
- [ ] *Θ*(log *N*)
- [ ] *Θ*(*N*)
{% endcapture %}
{% include question.html question=question number=1 %}
