---
title: Tree rotation
nav_order: 0
---

2-3 trees are complicated to implement because overstuffing and splitting requires frequently constructing new 2-nodes and 3-nodes. In practice, creating new nodes so often can slow down the data structure even though it is (generally) an asymptotic improvement over binary search trees. In this lesson, we will learn about a self-balancing **binary search tree** that has a one-to-one correspondence with 2-3 trees.

Depending on the order items are inserted into the tree, there are many different ways to construct a binary search tree containing the same set of items. However, insertion order is not the only way to build different configurations of the same set of items. We can change the structure of a binary tree through **rotation**.

Each rotation makes a local adjustment to the tree, sometimes increasing and sometimes decreasing the height of the tree but always respecting the binary search tree invariant. Items that are ordered between **B** and **D** in the example below stay ordered after a rotation.

![Rotation Respects Invariants](https://docs.google.com/drawings/d/e/2PACX-1vSxtNc8LBeAA_0lfsjznw8V0m3NpLnY8wdxFcBHzpoeTs4h-INPcrHg9SrTxiBE45LnpVAuaVxbgLqX/pub?w=960&h=540)

Given a particular node in a tree, we can either rotate it to the left or to the right by making it the child of one of its parents. In these definitions, **G** and **P** correspond to their labels in the tree shown below.

rotateLeft(**G**)
: Let `x` be the right child of **G**. Make **G** the new left child of `x`. We can think of this as temporarily merging **G** and **P**, then sending **G** down and left.

  ![Rotate Left](https://docs.google.com/drawings/d/e/2PACX-1vTuNPs88LYDo0zqxw7CaWp4Kd8lToIon_VqAS6PKySXsPh1bzjzI-ypJr7KaxQwX7c0o7UTIv547OoT/pub?w=960&h=540)

rotateRight(**P**)
: Let `x` be the left child of **P**. Make **P** the new right child of `x`. We can think of this as temporarily merging **G** and **P**, then sending **P** down and right.

  ![Rotate Right](https://docs.google.com/drawings/d/e/2PACX-1vRQzETsHHN79oS1sh9H9mtD5Ct9_5DIXS0bVD1e-PmY5RjcqNyi32i3uon-EJQLc2D0vect6SpHpevx/pub?w=960&h=540)

{% capture question %}
Give a sequence of rotation operations that balances the tree on the left.

![Rotation Goal](https://docs.google.com/drawings/d/e/2PACX-1vQmjAzDQAzKQkbEC4cN8Ur2Up1Awvmdfn-CTexDCR7gbvwHCdRjkh6Glgro3CRXxnJEz3ILLffn0hRF/pub?w=960&h=270)
{% endcapture %}
{% capture explanation %}
There are many possible answers!

![Rotation Operations](https://docs.google.com/drawings/d/e/2PACX-1vR73xxplcCcEUIbtoX8aWEDGvyCdngGvnBz8Lxtk1Ttt_FV1kiyqTUefxhCwhfrSRJcivgmxh-RX_3y/pub?w=960&h=270)
{% endcapture %}
{% include question.html question=question explanation=explanation number=1 %}

{% capture question %}
After performing a rotation, which of the following can occur?

- [ ] If the tree had the search tree property, it may lose it.
- [ ] If the tree did not have the search tree property, it may gain it.
- [ ] The height can stay the same.
- [ ] The height can increase.
- [ ] The height can decrease.
- [ ] The number of nodes can change.
- [ ] The root can change.
{% endcapture %}
{% include question.html question=question number=2 %}
